/* vi: set sw=2 ts=8: */
/*
 * Copyright 2005 by HI Corporation.
 *
 * This source code may be used for any purpose as long as this
 * copyright message remains in the source code and the documentation
 * for any product made from it includes a note stating:
 *
 * 	"Incorporates software from HI Corporation, copyright 2005."
 * 
 * This code comes WITHOUT warranty of merchantability or fitness for
 * a particular purpose.
 */

/* Sample Program accompanying the Using M3G section of Siggraph 2005
 * course #35.
 */

import java.lang.Math;
import javax.microedition.m3g.*;

/**
 * A camera manipulator. This class applies rotations to a World's
 * activeCamera that make it rotate around the prime axes passing
 * through the World's origin.
 */
public class CameraManip
{
  Camera curCamera;
  float distToTarget;
  float x, y, z, dx, dy, dz;
  float angleX, angleY, angleZ;
  Transform curOriginalXform;
  Transform cameraXform, transXform, rotateXform;

  public CameraManip(World world)
  {
    Transform world2Cam = new Transform();
    float[] matrix = new float[16];

    curOriginalXform = new Transform();
    cameraXform = new Transform();
    transXform = new Transform();
    rotateXform = new Transform();
    x = y = z = dx = dy = dz = 0;
    angleX = angleY = angleZ = 0;

    curCamera = world.getActiveCamera();
    if (curCamera != null) {
      curCamera.getTransformTo( world, world2Cam );
      world2Cam.get( matrix );
      distToTarget = (float)Math.sqrt( matrix[3]*matrix[3]
	  		      	       + matrix[7]*matrix[7]
			               + matrix[11]*matrix[11] );
 
      curCamera.getTransform( curOriginalXform );

      rotate( 0, 0, 0 );
      world2Cam = null;
    }
  }

  public void buildCameraXform() 
  {
    cameraXform.setIdentity();
    rotateXform.setIdentity();
    transXform.setIdentity();
    
    transXform.postTranslate( x + dx, y + dy, z + dz );
     
    // rotate about the x-axis then the y-axis
    rotateXform.postRotate( angleY, 0, 1, 0 );
    rotateXform.postRotate( angleX, 1, 0, 0 );

    cameraXform.postMultiply( transXform );
    cameraXform.postMultiply( rotateXform );
  }


  public void baseRotate(float dAngleX, float dAngleY, float dAngleZ)
  {
    angleX += dAngleX;
    if( Math.abs( angleX ) >= 360 ) {
      angleX -= 360;
    }

    angleY += dAngleY;
    if( Math.abs( angleY ) >= 360 ) {
      angleY -= 360;
    }

    angleZ += dAngleZ;
    if( Math.abs( angleZ ) >= 360 ) {
      angleZ -= 360;
    }
  }


  public void rotate(float dAngleX, float dAngleY, float dAngleZ)
  {
    if (curCamera == null)
      return;
    
    baseRotate( dAngleX, dAngleY, dAngleZ );

    /** adjust the camera position. */
    Transform rotTrans = new Transform(); 
    /** new transforms are set to identity */

    rotTrans.postRotate( angleY, 0, 1, 0 );
    rotTrans.postRotate( angleX, 1, 0, 0 );

    float pos[] = { 0, 0, distToTarget, 1 };

    rotTrans.transform( pos );
    dx = pos[0]; 
    dy = pos[1]; 
    dz = pos[2] - distToTarget;

    buildCameraXform();
    setCameraXform();
    rotTrans = null;
  }


  public void setCameraXform() 
  {
    cameraXform.postMultiply( curOriginalXform );
    curCamera.setTransform( cameraXform );
  }

}


